from ..crazy import CrazyDragon

from threading import Thread

from cflib.crazyflie.syncCrazyflie import SyncCrazyflie

from time import sleep



class SwarmHandler:

    @staticmethod
    def OPENLINK( _uris: dict ):

        _cfs = {}

        _scfs = {}

        for _uri, bodyname in _uris.items():
            
            print( "<<< ", _uri, " >>>" )

            _cf  = CrazyDragon()
            _scf = SyncCrazyflie( _uri, cf=_cf )
        
            _cfs[bodyname]  = _cf
            _scfs[bodyname] = _scf

            _scf.open_link()

            sleep( 2 )

        return _cfs, _scfs

    
    @staticmethod
    def CLOSELINK( _scfs: dict ):

        for _, _scf in _scfs.items():
            _scf.close_link()


    def parallel_run( self, func, _cfs: dict, args_list: list=None, args_dict: dict=None ):

        threads = []

        for bodyname, _cf in _cfs.items():

            args = [ _cf ]

            if ( args_list != None ):
                args += args_list

            if ( args_dict != None ):
                args += args_dict[bodyname]

            thread = Thread( target=func, args=args )
            threads.append( thread )
            thread.start()

        for thread in threads:
            
            thread.join()


from ...crazy import CrazyDragon

from threading import Thread

from ..swarm_handler import SwarmHandler

from ..._base._navigation_base.imu       import IMU
from ..._base._navigation_base.imu_setup import preflight_sequence
from ..._base._navigation_base.qualisys  import Qualisys

from time import sleep



class SwarmNavigation( Thread, SwarmHandler ):

    def __init__( self, _cfs: dict ):

        super().__init__()

        self.daemon = True

        self._cfs = _cfs

        self.imu = {}

        for bodyname, _cf in _cfs.items():
            self.imu[bodyname] = IMU( _cf )

        self.qtm         = Qualisys( _cfs )
        self.qtm.on_pose = {}

        self.navigate = True
        self.doing    = True


    @classmethod
    def _on_pose( cls, cf: CrazyDragon, data: list ):

        cf.pos[:] = data[0:3]
        cf.att[:] = data[3:6]

        cf.extpos.send_extpos( data[0], data[1], data[2] )


    def run( self ):

        _cfs = self._cfs

        imu = self.imu
        qtm = self.qtm

        print( 'preflight sequence' )

        self.parallel_run( preflight_sequence, _cfs )

        for bodyname, _imu in imu.items():
            
            _imu.start_get_acc()
            _imu.start_get_vel()

        for bodyname, _cf in _cfs.items():

            qtm.on_pose[bodyname] = lambda pose: __class__._on_pose( _cf, pose )

        print( 'false' )
        
        self.doing = False

        while self.navigate:

            sleep( 0.1 )


    def wait_until_init( self ):

        while self.doing:

            sleep( 0.1 )


    def join(self):

        self.navigate = False

        super().join()


from ...crazy import CrazyDragon

from .imu_setup import *

from cflib.crazyflie.log import LogConfig



period_in_ms = 10


class IMU:

    def __init__( self, cf: CrazyDragon ):

        self.cf = cf


    def start_get_vel( self, period_in_ms=period_in_ms ):
        log_conf = LogConfig( name="velocity", period_in_ms=period_in_ms )
        log_conf.add_variable( 'stateEstimate.vx', 'FP16' )
        log_conf.add_variable( 'stateEstimate.vy', 'FP16' )
        log_conf.add_variable( 'stateEstimate.vz', 'FP16' )

        self.cf.log.add_config( log_conf )
        log_conf.data_received_cb.add_callback( self.velocity_callback )
        log_conf.start()
 

    def start_get_acc(self, period_in_ms=period_in_ms):
        log_conf = LogConfig( name='acceleration', period_in_ms=period_in_ms )
        log_conf.add_variable( 'acc.x', 'FP16' )      ## m/s^2
        log_conf.add_variable( 'acc.y', 'FP16' )      ## m/s^2
        log_conf.add_variable( 'acc.z', 'FP16' )      ## m/s^2

        self.cf.log.add_config( log_conf )
        log_conf.data_received_cb.add_callback( self.accelerate_callback )
        log_conf.start()


    def velocity_callback( self, timestamp, data, logconf ):
        self.cf.vel[0] = data['stateEstimate.vx']
        self.cf.vel[1] = data['stateEstimate.vy']
        self.cf.vel[2] = data['stateEstimate.vz']


    def accelerate_callback( self, timestamp, data, logconf ):
        self.cf.acc[0] = data['acc.x'] * 9.81
        self.cf.acc[1] = data['acc.y'] * 9.81
        self.cf.acc[2] = data['acc.z'] * 9.81


from setuptools import setup, find_packages



setup(
    name='CrazyDG',
    version='5.0.4',
    description='CrazyDG is for convenience',
    author='Daegeun02',
    author_email='redhawkdg02@gmail.com',
    url='https://github.com/Daegeun02/CrazyDG.git',
    install_requires=[],
    packages=find_packages( exclude=[] ),
    keywords=['CrazyDG', 'by daegeun', 'for convenience'],
    python_requires='>=3.7',
    package_data={},
    zip_safe=False,
    classifiers=[
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
)



from ...crazy import CrazyDragon

from .constants import Kp, Kd, g

from numpy import array, zeros

from time import sleep



def takeoff( cf: CrazyDragon, h=1.5, T=3, dt=0.1 ):

    cur     = zeros(3)
    des     = zeros(3)
    acc_cmd = zeros(3)
    P_pos   = zeros(3)
    D_pos   = zeros(3)
    care_g  = array([0,0,g])

    print( 'take-off' )

    n = int( T / dt )
    t = 0

    command = cf.command

    cur[:] = cf.pos
    pos    = cf.pos
    vel    = cf.vel

    des[ 0 ] = cur[0]
    des[ 1 ] = cur[1]
    des[ 2 ] = h

    cf.destination[:] = des

    for _ in range( n ):

        P_pos[:] = des - pos
        D_pos[:] = vel

        acc_cmd[:] = 0
        acc_cmd[:] += P_pos * Kp
        acc_cmd[:] -= D_pos * Kd
        acc_cmd[:] += care_g

        command[:] = acc_cmd

        t += dt

        sleep( dt )


from ..crazy import CrazyDragon

from threading import Thread

from .integrator import _RK4
from .integrator import _A_RK4

from numpy import zeros
from numpy import ndarray


from time import sleep



class Dynamic4PILS( Thread ):

    def __init__( self, cf: CrazyDragon, dt ):

        super().__init__()

        self._cf = cf
        self._dt = dt

        self.dxdt = zeros((6,9))

        self._build_dynamic( dt )

        self.propagate = True

    
    def _build_dynamic( self, dt ):

        dxdt = self.dxdt

        dxdt[0,0] = 1
        dxdt[1,1] = 1
        dxdt[2,2] = 1
        dxdt[3,3] = 1
        dxdt[4,4] = 1
        dxdt[5,5] = 1
        dxdt[0,3] = dt
        dxdt[1,4] = dt
        dxdt[2,5] = dt

        dxdt[0,6] = 0.5 * dt * dt
        dxdt[1,7] = 0.5 * dt * dt
        dxdt[2,8] = 0.5 * dt * dt
        dxdt[3,6] = dt
        dxdt[4,7] = dt
        dxdt[5,8] = dt

    
    def deriv_x( cf: CrazyDragon, out: ndarray ):

        if( out.size != 6 ):
            raise ValueError( 'wrong size' )

        def _dxdt( t, x, args ):

            out[0:3] = x[3:6]
            out[3:6] = cf.command

            return out

        return _dxdt


    def run( self ):

        cf   = self._cf
        dt   = self._dt

        pos = cf.pos
        vel = cf.vel

        x    = zeros(6)
        xdot = zeros(6)
        dxdt = self.deriv_x( cf, xdot )

        t = 0

        x[0:3] = pos
        x[3:6] = vel

        while self.propagate:
            _RK4( dxdt, t, x, dt )

            pos[:] = x[0:3]
            vel[:] = x[3:6]            

            t += dt

            sleep( dt )


from ...crazy import CrazyDragon

from time import sleep

from cflib.crazyflie.log        import LogConfig
from cflib.crazyflie.syncLogger import SyncLogger



orientation_std_dev = 8.0e-3
SEND_FULL_POSE      = True


def preflight_sequence( _cf: CrazyDragon ):
    """
    This is the preflight sequence. It calls all other subroutines before takeoff.
    """

    adjust_orientation_sensitivity( _cf )

    activate_kalman_estimator( _cf )

   # enable high level commander
    _cf.param.set_value('commander.enHighLevel', '1')

    # prepare for motor shut-off
    _cf.param.set_value('motorPowerSet.enable', '0')
    _cf.param.set_value('motorPowerSet.m1', '0')
    _cf.param.set_value('motorPowerSet.m2', '0')
    _cf.param.set_value('motorPowerSet.m3', '0')
    _cf.param.set_value('motorPowerSet.m4', '0')

    print( 'upload' )

    # ensure params are downloaded
    _wait_for_param_download( _cf )

    print( 'upload done' )

    # reset the estimator
    _reset_estimator( _cf )

    print( 'reset_done' )

    # check state
    _check_state( _cf )

    print( 'sensors is on all green' )


def _wait_for_param_download( _cf: CrazyDragon ):

    while not _cf.param.is_updated: 
        
        print( 'waiting' )

        sleep(1.0)
 

def adjust_orientation_sensitivity( cf: CrazyDragon ):
    cf.param.set_value('locSrv.extQuatStdDev', orientation_std_dev)


def activate_kalman_estimator( cf: CrazyDragon ):
    cf.param.set_value('stabilizer.estimator', '2')

    # Set the std deviation for the quaternion data pushed into the
    # kalman filter. The default value seems to be a bit too low.
    cf.param.set_value('locSrv.extQuatStdDev', 0.06)


def _reset_estimator( cf: CrazyDragon ):

    cf.param.set_value('kalman.resetEstimation', '1')

    sleep(0.1)

    cf.param.set_value('kalman.resetEstimation', '0')

    # time.sleep(1)
    wait_for_position_estimator( cf )


def wait_for_position_estimator( cf: CrazyDragon ):
    print('Waiting for estimator to find position...')

    log_config = LogConfig(name='Kalman Variance', period_in_ms=10)
    log_config.add_variable('kalman.varPX', 'FP16')
    log_config.add_variable('kalman.varPY', 'FP16')
    log_config.add_variable('kalman.varPZ', 'FP16')

    var_y_history = [1000] * 10
    var_x_history = [1000] * 10
    var_z_history = [1000] * 10

    threshold = 0.001

    with SyncLogger( cf, log_config ) as logger:
        for log_entry in logger:
            data = log_entry[1]

            var_x_history.append( data['kalman.varPX'] )
            var_x_history.pop( 0 )
            var_y_history.append( data['kalman.varPY'] )
            var_y_history.pop( 0 )
            var_z_history.append( data['kalman.varPZ'] )
            var_z_history.pop( 0 )

            min_x = min( var_x_history )
            max_x = max( var_x_history )
            min_y = min( var_y_history )
            max_y = max( var_y_history )
            min_z = min( var_z_history )
            max_z = max( var_z_history )

            if ( max_x - min_x ) < threshold and (
                    max_y - min_y ) < threshold and (
                    max_z - min_z ) < threshold:
                break


def _check_state( _cf: CrazyDragon ):

    log_config = LogConfig(name='State', period_in_ms=500)
    log_config.add_variable('stabilizer.roll', 'float')
    log_config.add_variable('stabilizer.pitch', 'float')
    log_config.add_variable('stabilizer.yaw', 'float')

    with SyncLogger( _cf, log_config ) as sync_logger:

        for log_entry in sync_logger:

            log_data = log_entry[1]
            roll     = log_data['stabilizer.roll']
            pitch    = log_data['stabilizer.pitch']
            yaw      = log_data['stabilizer.yaw']

            #('yaw', yaw)
            if SEND_FULL_POSE:
                euler_checks = [('roll', roll, 5), ('pitch', pitch, 5)]
            else:
                euler_checks = [('roll', roll, 5), ('pitch', pitch, 5), ('yaw', yaw, 5)]

            for name, val, val_max in euler_checks:
                if abs(val) > val_max:
                    msg = "too much {:s}, {:10.4f} deg, for {:s}".format(
                        name, val, _cf.link_uri)
                    print(msg)
            return


from crazy      import CrazyDragon # >>> from CrazyDG import CrazyDragon
from navigation import Navigation  # >>> from CrazyDG import Navigation
from control    import Controller  # >>> from CrazyDG import Controller
from recorder   import Recorder    # >>> from CrazyDG import Recorder
from guidance   import utils       # >>> from CrazyDG import utils

from cflib                         import crtp
from cflib.crazyflie.syncCrazyflie import SyncCrazyflie
from cflib.utils                   import uri_helper

nav_config = {
    'body_name': 'cf1'
}

ctr_config = {
    'dt': 0.1,
    'n' : 5
}



uri = uri_helper.uri_from_env( default='radio://0/80/2M/E7E7E7E702' )


if __name__ == "__main__":

    crtp.init_drivers()
    
    _cf = CrazyDragon()

    with SyncCrazyflie( uri, cf=_cf ) as scf:

        NAV = Navigation( _cf, nav_config )
        CTR = Controller( _cf, ctr_config )
        RCD = Recorder( _cf, CTR )

        NAV.start()
        CTR.start()
        RCD.start()

        ## your guidance function ##
        CTR.init_send_setpoint()
        ##       from here        ##

        utils.takeoff( _cf )

        utils.hover( _cf, T=2 )

        utils.goto( _cf, [2.0,2.0,2.0], T=5 )

        utils.hover( _cf, T=2 )

        utils.landing_supporter( _cf, RCD )

        ############################

        CTR.stop_send_setpoint()

        NAV.join()
        CTR.join()
        RCD.join()

    NAV.qtm.close()
